from tradebot.requests import assert_ok_response, TradebotRequests
from tradebot.configs import TradebotConfigs


def __request_new_access_token(configs:TradebotConfigs):

        consumer_key, refresh_token = configs["consumer_key"], configs["refresh_token"]
        url = "https://api.tdameritrade.com/v1/oauth2/token"
        
        req_body = {
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "client_id": consumer_key + "@AMER.OAUTHAP",
        }

        requests = TradebotRequests(configs)
        res = requests.post(url, data=req_body, auth=False)
        assert_ok_response(res)

        return res.json()


def update_access_token(configs:TradebotConfigs):
    '''
    Requests a new access token and writes it to the config file.
    '''
    new_data = __request_new_access_token(configs)
    configs.update(new_data)



from typing import Dict
import requests
from requests import Response
from requests.exceptions import RequestException

from tradebot.configs import TradebotConfigs


class TradebotRequests:
    '''
    Performs the essential CRUD requests, using the ``access_token`` defined in the config file\n
    to perform Bearer authorization.
    '''
    def __init__(self, configs:TradebotConfigs):
        self.__configs = configs

        self.__headers = {
            "Accept": "*/*",
            "Authorization": None, #be sure to update Authorization each time the 'headers' are referenced
            "Content-Type": "application/x-www-form-urlencoded",
        }


    def __update_headers(self, auth:bool):
        auth_str = f"Bearer {self.__configs['access_token']}" if auth else None
        self.__headers.update({"Authorization": auth_str})


    def get(self, url:str, data:Dict=None, params:Dict=None, auth:bool=True):
        '''
        Uses the ``access_token`` defined in the configs file to make a Bearer authorized request if ``auth`` is True.
        '''
        self.__update_headers(auth)

        return requests.get(url, data=data, params=params, headers=self.__headers)


    def post(self, url:str, data:Dict=None, params:Dict=None, auth:bool=True):
        '''
        Uses the ``access_token`` defined in the configs file to make a Bearer authorized request if ``auth`` is True.
        '''
        self.__update_headers(auth)

        return requests.post(url, data=data, params=params, headers=self.__headers)


def assert_ok_response(res:Response):
        if not res.ok:
            raise RequestException(f"The response status code was {res.status_code}", response=res)



import json
from pathlib import Path
from typing import Dict


class TradebotConfigs:
    
    def __init__(self, file:Path):
        self.__file = file


    def get(self, key:str):
        return self[key]


    def set(self, key:str, val:str):
        self[key] = val


    def update(self, mapping:Dict):
        with open(self.__file) as f:
            data:dict = json.load(f)

        data.update(mapping)

        with open(self.__file, 'w') as f:
            json.dump(data, f, indent=2)

    
    def __getitem__(self, key:str):
        with open(self.__file) as f:
            data:dict = json.load(f)

        return data.get(key)


    def __setitem__(self, key:str, val:str):
        with open(self.__file) as f:
            data:dict = json.load(f)
        
        data[key] = val

        with open(self.__file, 'w') as f:
            json.dump(data, f, indent=2)



from tradebot.configs import TradebotConfigs
from tradebot.requests import TradebotRequests, assert_ok_response
from tradebot.backend.enums import CandleFrequencyType, CandlePeriodType

from bs4 import BeautifulSoup as BS
import requests


class Tradebot:
    '''
    Performs all the essential functionality related to stocks/securities, as defined by the TDAmeritrade API.\n
    '''

    def __init__(self, configs:TradebotConfigs):
        self.__configs = configs
        self.__tb_requests = TradebotRequests(configs)


    def fundamentals(self, ticker:str):
        '''
        Returns json containing fundamental data for the given ``ticker``.
        '''
        url = "https://api.tdameritrade.com/v1/instruments"
        params = {"apikey": self.__configs["consumer_key"], "symbol": ticker, "projection": "fundamental"}
        
        res = self.__tb_requests.get(url, params=params)
        assert_ok_response(res)

        return res.json()
        

    def quote(self, ticker:str):
        '''
        Returns json representing the quote for the given ``ticker``.
        '''
        url = f"https://api.tdameritrade.com/v1/marketdata/{ticker}/quotes"
        res = self.__tb_requests.get(url)
        assert_ok_response(res)

        return res.json()


    def price_history(
        self,
        ticker:str,
        period_type:CandlePeriodType=None,
        period:int=None,
        frequency_type:CandleFrequencyType=None,
        frequency:int=None,
        end_date:int=None,
        start_date:int=None,
        need_extended_hours:bool=True
    ):
        '''
        Returns the price history for the given ``ticker`` using the conditions defined by the following params:\n
        ``period_type``: The type of period to show. Valid values are 'day', 'month', 'year', or 'ytd' (year to date).\n
        ``period``: number of periods of ``period_type`` to show.\n
        Valid periods by periodType:
            day: 1, 2, 3, 4, 5, 10
            month: 1, 2, 3, 6
            year: 1, 2, 3, 5, 10, 15, 20
            ytd: 1
        ``frequency_type``: The type of frequency with which a new candle is formed.\n
        Valid frequencyTypes by periodType:
            day: minute
            month: daily, weekly
            year: daily, weekly, monthly
            ytd: daily, weekly
        ``frequency``: The number of the frequencyType to be included in each candle.\n
        Valid frequencies by frequencyType:
            minute: 1, 5, 10, 15, 30
            daily: 1
            weekly: 1
            monthly: 1
        '''
        url = f"https://api.tdameritrade.com/v1/marketdata/{ticker}/pricehistory"
        api_key = self.__configs["consumer_key"]
        params = {
            "apikey": api_key,
            "periodType": period_type._value_,
            "period": period,
            "frequencyType": frequency_type._value_,
            "frequencey": frequency,
            "endDate": end_date,
            "startDate": start_date,
            "needExtendedHoursData": need_extended_hours
        }

        res = self.__tb_requests.get(url, params=params)
        assert_ok_response(res)

        return res.json()



from enum import Enum


class CandleFrequencyType(Enum):
    MINUTE = "minute"
    DAILY = "daily"
    WEEKLY = "weekly"
    MONTHLY = "monthly"


class CandlePeriodType(Enum):
    DAY = "day"
    MONTH = "month"
    YEAR = "year"
    YTD = "ytd"


from typing import Dict, List
from tradebot.requests import TradebotRequests, assert_ok_response
from tradebot.configs import TradebotConfigs


def get_account_info(configs:TradebotConfigs, fields:str=None):
    acct_number = configs["account_number"]
    url = f"https://api.tdameritrade.com/v1/accounts/{acct_number}"
    url_params = {"fields": fields} if fields else ''
    requests = TradebotRequests(configs)
    res = requests.get(url, params=url_params)
    assert_ok_response(res)

    return res.json()


def get_balances(configs:TradebotConfigs):
    '''
    Returns a Dictionary containing balances for the account defined by ``account_number`` in the config file.
    '''
    acct_info = get_account_info(configs)
    sec_acct = acct_info.get("securitiesAccount", dict())
    balances = {
        "currentBalances": sec_acct.get("currentBalances", None),
        "initialBalances": sec_acct.get("initialBalances", None),
        "projectedBalances": sec_acct.get("projectedBalances", None)
    }
    return balances


def get_positions(configs:TradebotConfigs) -> List[Dict] | None:
    '''
    Returns a list of dictionaries, each representing a position in a particular security.\n
    Returns None if the user doesn't have any positions.
    '''
    acct_info = get_account_info(configs, "positions")
    positions = acct_info.get("securitiesAccount", dict()).get("positions", None)
    return positions


def get_orders(configs:TradebotConfigs) -> List[Dict] | None:
    '''
    Returns a list of dictionaries, each representing an order.
    '''
    acct_info = get_account_info(configs, "orders")
    orders = acct_info.get("securitiesAccount", dict()).get("orderStrategies", None)
    return orders